{$i-}{$r+}{$q+}{$s+}
uses crt;
const
 DELAY_PERIOD = 20;

 MAX_SIZE = 40;
 MAX_X = 40;
 MAX_Y = 20;

 EMPTY = 0;
 WALL = 1;
 CARROT = 2;

 WAS_VISITED = 1;
 NOT_VISITED = -1;

 INPUT_FILE = 'madu.sis';
 OUTPUT_FILE = 'madu.val';

 BAD_INPUT_STR = 'Bad input!';
 BAD_OUTPUT_STR = 'Bad output!';

 WRONG_START_STR = 'Wrong start';
 IN_THE_WALL_STR = 'In the wall';
 OUT_OF_CAVE_STR = 'Out of the cave';
 SELF_EAT_STR = 'Eats itself';
 WRONG_END_STR = 'Wrong end';
 CARROTS_NOT_EATEN_STR = 'Not all carrots have been eaten';
 SOLUTION_CORRECT_STR = 'Solution correct';
 PRESS_ENTER_STR = 'Press ENTER to continue...';

 INTERNAL_ERROR_STR = 'Internal error';
 PROGRAM_HALTED_STR = 'Program halted';

 BODY_CHAR = ' ';
 HEAD_CHAR = #1;
 DEAD_CHAR = #2;

type
 map = array[1..MAX_SIZE,1..MAX_SIZE] of integer;
 list = array[1..MAX_SIZE*MAX_SIZE] of record
                                        dx,dy:shortint;
                                       end;

var
 isAutomatic:boolean;

 m,n:integer;
 cave:map;
 visited:map;

 sol_start_x, sol_start_y:integer;
 sol_len:integer;
 solution:list;

 solutionCorrect:boolean;
 solutionVerdict:String;


procedure waitEnter;
begin
 gotoxy(1,24);
 textcolor(white);
 textbackground(red);

 writeln(PRESS_ENTER_STR);
 while readkey <> #13 do;

 textcolor(white);
 textbackground(black);
end;


procedure raiseError(message:String);
begin
 gotoxy(1,23);
 textcolor(white);
 textbackground(red);

 writeln('Fatal error:', message);

 waitEnter;

 halt(0);
end;

procedure read_input;
var
 f:text;
 ch:char;
 x,y:integer;
begin
 assign(f,INPUT_FILE);
 reset(f);
 if ioresult <> 0 then raiseError(BAD_INPUT_STR);
 readln(f,m,n);
 if ioresult <> 0 then raiseError(BAD_INPUT_STR);

 if (m < 3) or (n < 3) or
    (m > MAX_X) or (n > MAX_Y) then raiseError(BAD_INPUT_STR);

 for y:=1 to n do
 begin
  for x:=1 to m do
  begin
   read(f,ch);
   if ioresult <> 0 then raiseError(BAD_INPUT_STR);

   case ch of
    '#':cave[x,y]:=WALL;
    '!':cave[x,y]:=CARROT;
    '.':cave[x,y]:=EMPTY;
   else
    raiseError(BAD_INPUT_STR);
   end;
  end;
  readln(f);
  if ioresult <> 0 then raiseError(BAD_INPUT_STR);
 end;

 close(f);
end;

procedure read_output;
var
 f:text;
 ch:char;
begin
 assign(f,OUTPUT_FILE);
 reset(f);
 if ioresult <> 0 then raiseError(BAD_OUTPUT_STR);

 readln(f,sol_start_x, sol_start_y);
 if ioresult <> 0 then raiseError(BAD_OUTPUT_STR);

 sol_start_y:= 1+n-sol_start_y;

 sol_len:=0;
 while not eoln(f) do
 begin
  read(f,ch);
  if ioresult <> 0 then raiseError(BAD_OUTPUT_STR);

  inc(sol_len);

  case ch of
   'N': begin
         solution[sol_len].dx:=0;
         solution[sol_len].dy:=-1;
        end;
   'S': begin
         solution[sol_len].dx:=0;
         solution[sol_len].dy:=1;
        end;
   'E': begin
         solution[sol_len].dx:=1;
         solution[sol_len].dy:=0;
        end;
   'W': begin
         solution[sol_len].dx:=-1;
         solution[sol_len].dy:=0;
        end;
   else
    raiseError(BAD_OUTPUT_STR);
   end;
 end;

 close(f);
end;

function verifyPoint(x,y:integer):boolean;
begin
 verifyPoint:=(x >= 1) and (x <= m) and (y >=1) and (y <= n);
end;

function symbol(i:integer):char;
begin
 case i of
  WALL: symbol:='#';
  CARROT: symbol:='!';
  EMPTY: symbol:='.';
 else
  raiseError(INTERNAL_ERROR_STR);
 end;
end;

function on_edge(x,y:integer):boolean;
begin
 on_edge:=(x=1) or (x=m) or (y=1)or(y=n);
end;

procedure drawSegment(x,y:integer;c:char);
begin
 gotoxy(x,y);
 textBackground(blue);
 textColor(yellow);
 write(c);
end;

procedure check_solution;
var
 carrot_n:integer;
 x,y,c:integer;
 l:integer;
begin
 solutionCorrect:=false;

 carrot_n:=0;

 for y:=1 to n do
 begin
  for x:=1 to m do
  begin
   visited[x,y]:= NOT_VISITED;
   if cave[x,y]=CARROT then inc(carrot_n);

   write(symbol(cave[x,y]));
  end;
  writeln;
 end;

 x:=sol_start_x;
 y:=sol_start_y;
 c:=0;

 if not verifyPoint(x,y) then
 begin
  solutionVerdict:=WRONG_START_STR;
  exit;
 end;

 if not on_edge(x,y) then
 begin
  drawSegment(x,y,DEAD_CHAR);
  solutionVerdict:=WRONG_START_STR;
  exit;
 end;

 if cave[x,y] = WALL then
 begin
  drawSegment(x,y,DEAD_CHAR);
  solutionVerdict:=IN_THE_WALL_STR;
  exit;
 end;

 visited[x,y] := WAS_VISITED;
 if cave[x,y] = CARROT then inc(c);

 drawSegment(x,y,HEAD_CHAR);

 for l:=1 to sol_len do
 begin
  if isAutomatic then
  begin
   delay(DELAY_PERIOD);
   if keypressed and (readkey=#27) then raiseError(PROGRAM_HALTED_STR);
  end
  else
   if readkey = #27 then raiseError(PROGRAM_HALTED_STR);

  drawSegment(x,y,BODY_CHAR);

  x:=x+solution[l].dx;
  y:=y+solution[l].dy;

  if not verifyPoint(x,y) then
  begin
   solutionVerdict:=OUT_OF_CAVE_STR;
   exit;
  end;

  if cave[x,y] = WALL then
  begin
   drawSegment(x,y,DEAD_CHAR);
   solutionVerdict:=IN_THE_WALL_STR;
   exit;
  end;

  if visited[x,y] = WAS_VISITED then
  begin
   drawSegment(x,y,DEAD_CHAR);
   solutionVerdict:=SELF_EAT_STR;
   exit;
  end;

  visited[x,y] := WAS_VISITED;
  if cave[x,y] = CARROT then inc(c);

  drawSegment(x,y,HEAD_CHAR);
 end;

 if not on_edge(x,y) then
 begin
  drawSegment(x,y,DEAD_CHAR);
  solutionVerdict:=WRONG_END_STR;
  exit;
 end;

 if not(c=carrot_n) then
 begin
  drawSegment(x,y,DEAD_CHAR);
  solutionVerdict:=CARROTS_NOT_EATEN_STR;
  exit;
 end;

 solutionVerdict:=SOLUTION_CORRECT_STR;
 solutionCorrect:=true;
end;

procedure drawVerdict;
begin
 gotoxy(1,23);
 textcolor(white);
 textbackground(red);
 writeln(solutionVerdict);
end;

begin
 textcolor(white);
 textbackground(black);
 clrscr;

 isAutomatic:=true;
 if paramcount > 0 then
   if paramstr(1) = '-k' then isAutomatic:= false;

 read_input;
 read_output;
 check_solution;
 drawVerdict;

 waitEnter;
end.